import datetime
import io
import json
import logging
import os.path
import re

import click
import httpx
import humanize
import jsonslicer
import psycopg
import tornado.ioloop
import tornado.web
import tqdm

import tutor.csvimport
import tutor.database
import tutor.scryfall
import tutor.search
import tutor.server


@click.group()
@click.option(
    "--database",
    envvar="TUTOR_DATABASE",
    type=click.Path(dir_okay=False),
    required=True,
)
@click.option(
    "--log-level",
    type=click.Choice(
        ["debug", "info", "warn", "error"],
        case_sensitive=False,
    ),
    default="warn",
)
@click.pass_context
def main(ctx, database, log_level):
    logging.basicConfig(
        level={
            "debug": logging.DEBUG,
            "info": logging.INFO,
            "warn": logging.WARN,
            "error": logging.ERROR,
        }.get(log_level.lower())
    )
    ctx.ensure_object(dict)
    ctx.obj["database"] = database


@main.command()
@click.option("--port", type=int, envvar="TUTOR_PORT", default=8888)
@click.option("--scheme", envvar="TUTOR_SCHEME")
@click.option("--static", envvar="TUTOR_STATIC", type=click.Path(file_okay=False))
@click.option("--debug", is_flag=True)
@click.option("--reload", is_flag=True)
@click.pass_context
def server(ctx, port, scheme, static, debug, reload):
    app = tutor.server.Application(
        **{
            **ctx.obj,
            "scheme": scheme,
            "static": static,
            "debug": debug,
            "autoreload": reload,
        }
    )
    app.listen(port)
    tornado.ioloop.IOLoop.current().start()


@main.group("import")
def import_group(): ...


@import_group.command("cards")
@click.argument("filename", type=click.Path(dir_okay=False))
@click.option(
    "--mode",
    type=click.Choice([mode.value for mode in tutor.csvimport.ImportMode]),
    default=tutor.csvimport.ImportMode.add.value,
)
@click.pass_context
def import_cards(ctx, filename, mode):
    async def _import_cards():
        # count total lines
        with open(filename, "r") as csvfile:
            total = sum(1 for _ in csvfile)
        cards = 0
        with open(filename, "r") as csvfile:
            last = 0
            with tqdm.tqdm(total=total, unit="l") as bar:
                async for card, line in tutor.csvimport.load(
                    ctx.obj, csvfile, mode=tutor.csvimport.ImportMode[mode]
                ):
                    cards += 1
                    read = line - last
                    last = line
                    bar.set_description(f"{cards} cards imported")
                    bar.update(read)

    tornado.ioloop.IOLoop.current().run_sync(_import_cards)


@import_group.command("deck")
@click.argument("filename", type=click.Path(dir_okay=False))
@click.option("--name")
@click.pass_context
def import_deck(ctx, filename, name):
    line_pattern = re.compile(r"^(?P<quantity>\d+)x? (?P<name>[^\]$]+)")

    with open(filename) as f:
        lines = f.readlines()
    name = name or os.path.splitext(os.path.basename(filename))[0]

    async def import_deck():
        async with await psycopg.AsyncConnection.connect(
            ctx.obj["database"], autocommit=False
        ) as conn:
            async with conn.cursor() as cursor:
                deck_id = await tutor.database.store_deck(cursor, name)
                with tqdm.tqdm(lines) as bar:
                    bar.write(f"Importing deck '{name}' [{deck_id}]")
                    for line in bar:
                        if match := line_pattern.match(line.strip()):
                            groups = match.groupdict()
                            oracle_id = await tutor.database.oracle_id_by_name(
                                cursor, groups["name"]
                            )
                            if oracle_id:
                                await tutor.database.store_deck_card(
                                    cursor,
                                    deck_id,
                                    oracle_id,
                                    int(groups["quantity"]),
                                )
            await conn.commit()

    tornado.ioloop.IOLoop.current().run_sync(import_deck)


@main.command("update_scryfall")
@click.option("--filename", type=click.Path(dir_okay=False))
@click.pass_context
def update_scryfall(ctx, filename):
    class IterStreamer(object):
        """
        File-like streaming iterator.
        """

        def __init__(self, generator):
            self.generator = generator
            self.iterator = iter(generator)
            self.leftover = b""

        def __len__(self):
            return self.generator.__len__()

        def __iter__(self):
            return self.iterator

        def next(self):
            return next(self.iterator)

        def read(self, size):
            data = self.leftover
            count = len(self.leftover)

            if count < size:
                try:
                    while count < size:
                        chunk = self.next()
                        data += chunk
                        count += len(chunk)
                except StopIteration:
                    pass

            self.leftover = data[size:]

            return data[:size]

    class ScryfallFile:
        def __init__(self, filename):
            self.filename = filename
            self.total = os.path.getsize(filename)

        def cards(self):
            with open(self.filename) as f:
                data = jsonslicer.JsonSlicer(f, [None])
                last = f.tell()
                for card in data:
                    read = f.tell() - last
                    last = f.tell()
                    yield card, read

    class ScryfallAPI:
        def __init__(self):
            response = httpx.get("https://api.scryfall.com/bulk-data/all-cards")
            info = response.json()
            self.uri = info["download_uri"]
            self.total = info["compressed_size"]

        def cards(self):
            with httpx.stream("GET", self.uri) as response:
                downloaded = response.num_bytes_downloaded
                stream = IterStreamer(response.iter_bytes())
                data = jsonslicer.JsonSlicer(stream, [None])
                for card in data:
                    read = response.num_bytes_downloaded - downloaded
                    downloaded = response.num_bytes_downloaded
                    yield card, read

    if filename:
        source = ScryfallFile(filename)
    else:
        source = ScryfallAPI()

    async def import_cards():
        today = datetime.date.today()
        async with await psycopg.AsyncConnection.connect(
            ctx.obj["database"], autocommit=False
        ) as conn:
            async with conn.cursor() as cursor:
                with tqdm.tqdm(total=source.total, unit="b", unit_scale=True) as bar:
                    await cursor.execute(
                        """
                        CREATE TEMP TABLE "tmp_cards"
                        (LIKE "cards" INCLUDING DEFAULTS)
                        ON COMMIT DROP
                        """
                    )
                    await cursor.execute(
                        """
                        CREATE TEMP TABLE "tmp_prices"
                        (LIKE "card_prices" INCLUDING DEFAULTS)
                        ON COMMIT DROP
                        """
                    )
                    for card_json, read in source.cards():
                        try:
                            card = tutor.scryfall.to_card(card_json).unwrap()
                            await tutor.database.store_set(
                                cursor,
                                card_json["set"].upper(),
                                card_json["set_type"],
                                card_json["set_name"],
                            )
                            await tutor.database.store_card(cursor, card)
                            await tutor.database.store_price(cursor, today, card)
                        except Exception as e:
                            print(f"Skipping entry {card_json}: {e}")
                        bar.update(read)
                print("Updating oracle card data & indexes")
                await cursor.execute(
                    """
                    INSERT INTO "oracle"
                    SELECT DISTINCT ON("oracle_id")
                           "oracle_id"
                         , "name"
                         , "color_identity"
                         , "cmc"
                         , "mana_cost"
                         , "type_line"
                         , "edhrec_rank"
                         , "oracle_text"
                    FROM "tmp_cards"
                    WHERE "tmp_cards"."oracle_id" IS NOT NULL
                    ORDER BY "oracle_id"
                           , "name"
                           , "color_identity"
                           , "cmc"
                           , "mana_cost"
                           , "type_line"
                           , "edhrec_rank"
                           , "oracle_text"
                    ON CONFLICT (oracle_id) DO UPDATE
                    SET "name" = "excluded"."name"
                      , "color_identity" = "excluded"."color_identity"
                      , "cmc" = "excluded"."cmc"
                      , "mana_cost" = "excluded"."mana_cost"
                      , "type_line" = "excluded"."type_line"
                      , "edhrec_rank" = "excluded"."edhrec_rank"
                      , "oracle_text" = "excluded"."oracle_text"
                    """
                )
                print("Updating scryfall card data & indexes")
                await cursor.execute(
                    """
                    INSERT INTO "scryfall"
                         ( "scryfall_id"
                         , "oracle_id"
                         , "set_code"
                         , "collector_number"
                         , "release_date"
                         , "rarity"
                         , "artist"
                         )
                    SELECT "scryfall_id"
                         , "oracle_id"
                         , "set_code"
                         , "collector_number"
                         , "release_date"
                         , "rarity"
                         , "artist"
                    FROM "tmp_cards"
                    ON CONFLICT (scryfall_id) DO UPDATE
                    SET "oracle_id" = "excluded"."oracle_id"
                      , "set_code" = "excluded"."set_code"
                      , "collector_number" = "excluded"."collector_number"
                      , "release_date" = "excluded"."release_date"
                      , "rarity" = "excluded"."rarity"
                      , "artist"  = "excluded"."artist"
                    """
                )
                print("Updating card price data & indexes")
                await cursor.execute(
                    """
                    INSERT INTO "card_prices" SELECT * FROM "tmp_prices"
                    ON CONFLICT ("scryfall_id", "date") DO UPDATE
                    SET "usd" = "excluded"."usd"
                      , "usd_foil" = "excluded"."usd_foil"
                      , "eur" = "excluded"."eur"
                      , "eur_foil" = "excluded"."eur_foil"
                      , "tix" = "excluded"."tix"
                    """
                )
                print("Refreshing materialized views")
                await cursor.execute('REFRESH MATERIALIZED VIEW "cards"')
                await cursor.execute('REFRESH MATERIALIZED VIEW "oracle_latest"')

                await tutor.database.store_var(cursor, "last_update", str(today))
            await conn.commit()

    tornado.ioloop.IOLoop.current().run_sync(import_cards)


if __name__ == "__main__":
    main()
