import dataclasses
import enum
import functools
import typing

import parsy

import tutor.models


class Operator(enum.Enum):
    matches = ":"
    gte = ">="
    lte = "<="


@dataclasses.dataclass
class Criterion:
    operator: Operator


@dataclasses.dataclass
class Expansion(Criterion):
    set_code: parsy.string


@dataclasses.dataclass
class Color(Criterion):
    colors: typing.Set[tutor.models.Color]


@dataclasses.dataclass
class Rarity(Criterion):
    rarity: tutor.models.Rarity


@dataclasses.dataclass
class Type(Criterion):
    text: parsy.string


@dataclasses.dataclass
class Artist(Criterion):
    text: parsy.string


@dataclasses.dataclass
class Oracle(Criterion):
    text: parsy.string


@dataclasses.dataclass
class PriceUSD(Criterion):
    price: parsy.string


@dataclasses.dataclass
class Collection(Criterion):
    text: parsy.string


@dataclasses.dataclass
class Name(Criterion):
    operator: Operator = Operator.lte
    text: parsy.string = ""


@dataclasses.dataclass
class Search:
    criteria: typing.List[Criterion]


lstring = functools.partial(parsy.string, transform=lambda s: s.lower())
ustring = functools.partial(parsy.string, transform=lambda s: s.upper())
lstring_from = functools.partial(parsy.string_from, transform=lambda s: s.lower())
ustring_from = functools.partial(parsy.string_from, transform=lambda s: s.upper())

C, W, U, B, G, R = (
    tutor.models.Color.Colorless,
    tutor.models.Color.White,
    tutor.models.Color.Blue,
    tutor.models.Color.Black,
    tutor.models.Color.Green,
    tutor.models.Color.Red,
)

matches = parsy.string(":").map(Operator)
gte = parsy.string(">=").map(Operator)
lte = parsy.string("<=").map(Operator)

color = (
    ustring("c").result(C)
    | ustring("w").result(W)
    | ustring("u").result(U)
    | ustring("b").result(B)
    | ustring("g").result(G)
    | ustring("r").result(R)
)

multicolor = color.many().map(set)

single_color = (
    lstring("colorless").result({C})
    | lstring("white").result({W})
    | lstring("blue").result({U})
    | lstring("black").result({B})
    | lstring("green").result({G})
    | lstring("red").result({R})
)

guild = (
    lstring("boros").result({W, R})
    | lstring("golgari").result({G, B})
    | lstring("selesnya").result({G, W})
    | lstring("dimir").result({U, B})
    | lstring("orzhov").result({W, B})
    | lstring("izzet").result({U, R})
    | lstring("gruul").result({R, G})
    | lstring("azorius").result({W, U})
    | lstring("rakdos").result({B, R})
    | lstring("simic").result({G, U})
)

shard = (
    lstring("bant").result({W, G, U})
    | lstring("esper").result({U, W, B})
    | lstring("grixis").result({B, U, R})
    | lstring("jund").result({R, B, G})
    | lstring("naya").result({G, R, W})
)

wedge = (
    lstring("abzan").result({W, B, G})
    | lstring("jeskai").result({W, U, R})
    | lstring("sultai").result({U, B, G})
    | lstring("mardu").result({W, B, R})
    | lstring("temur").result({U, R, G})
)

any_color = single_color | guild | shard | wedge | multicolor

colors = parsy.seq(
    _keyword=lstring_from("c", "color", "colors"),
    operator=matches | gte | lte,
    colors=any_color,
).combine_dict(Color)

expansion_string = (
    parsy.regex(r"[a-zA-Z0-9]+").map(lambda s: s.upper()).desc("expansion set code")
)

expansion = parsy.seq(
    _keyword=lstring_from("e", "expansion", "s", "set"),
    operator=matches,
    set_code=expansion_string,
).combine_dict(Expansion)

# in_expansion = lstring("in") >> matches >> expansion_string.map(InExpansion)

any_rarity = (
    lstring_from("c", "common").result(tutor.models.Rarity.Common)
    | lstring_from("u", "uncommon").result(tutor.models.Rarity.Uncommon)
    | lstring_from("r", "rare").result(tutor.models.Rarity.Rare)
    | lstring_from("m", "mythic").result(tutor.models.Rarity.Mythic)
)

rarity = parsy.seq(
    _keyword=lstring_from("r", "rarity"),
    operator=matches | gte | lte,
    rarity=any_rarity,
).combine_dict(Rarity)

string_literal = parsy.regex(r'"[^"]*"').map(lambda s: s[1:-1]) | parsy.regex(r"[^\s]+")

type_line = parsy.seq(
    _keyword=lstring_from("t", "type"),
    operator=matches,
    text=string_literal,
).combine_dict(Type)

artist = parsy.seq(
    _keyword=lstring_from("a", "artist"),
    operator=matches,
    text=string_literal,
).combine_dict(Artist)

oracle = parsy.seq(
    _keyword=lstring_from("o", "oracle"),
    operator=matches,
    text=string_literal,
).combine_dict(Oracle)

collection = parsy.seq(
    _keyword=lstring_from("collection"),
    operator=matches,
    text=string_literal,
).combine_dict(Collection)

price_usd = parsy.seq(
    _keyword=lstring_from("usd"),
    operator=gte | lte,
    price=string_literal,
).combine_dict(PriceUSD)


exact_name = parsy.seq(_prefix=lstring_from("!"), text=string_literal).combine_dict(
    functools.partial(Name, operator=Operator.matches)
)
partial_name = parsy.seq(text=string_literal).combine_dict(Name)

name = exact_name | partial_name

criterion = (
    colors
    | expansion
    | rarity
    | type_line
    | artist
    | oracle
    | collection
    | price_usd
    | name
)

padding = parsy.regex(r"\s*")
search = padding >> (criterion << padding).many().map(Search)

if __name__ == "__main__":
    import sys

    print(search.parse(" ".join(sys.argv[1:])))
