import asyncio
import decimal
import importlib.metadata
import importlib.resources
import io
import itertools
import json
import logging
import re
import typing
import urllib.parse

import psycopg
import psycopg.rows
import psycopg_pool
import tornado.ioloop
import tornado.web
import tornado_openapi3.handler
import yaml
from tornado.ioloop import IOLoop

import tutor.database
import tutor.models
import tutor.search


logger = logging.getLogger(__name__)


def update_args(url: str, **qargs) -> str:
    parts = urllib.parse.urlsplit(url)
    return urllib.parse.urlunsplit(
        (
            parts.scheme,
            parts.netloc,
            parts.path,
            urllib.parse.urlencode(
                [
                    (k, v)
                    for k, v in urllib.parse.parse_qsl(parts.query)
                    if k not in qargs.keys()
                ]
                + list(qargs.items())
            ),
            parts.fragment,
        )
    )


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        if isinstance(obj, tutor.models.Card):
            return self._card(obj)
        if isinstance(obj, tutor.models.CardCopy):
            return self._card_copy(obj)
        if isinstance(obj, tutor.models.OracleCard):
            return self._oracle_card(obj)
        if isinstance(obj, tutor.models.Deck):
            return self._deck(obj)
        if isinstance(obj, tutor.models.DeckCard):
            return self._deck_card(obj)
        return json.JSONEncoder.default(self, obj)

    def _oracle_card(self, card: tutor.models.OracleCard) -> dict:
        return {
            "oracle_id": str(card.oracle_id),
            "name": card.name,
            "color_identity": tutor.models.Color.to_string(card.color_identity),
            "oracle_text": card.oracle_text,
            "cmc": float(card.cmc),
            "mana_cost": card.mana_cost,
            "type_line": card.type_line,
        }

    def _card(self, card: tutor.models.Card) -> dict:
        def price(amount: typing.Optional[decimal.Decimal]) -> typing.Optional[float]:
            if amount is not None:
                return float(amount)
            else:
                return None

        return {
            "scryfall_id": str(card.scryfall_id),
            "name": card.name,
            "set_code": card.set_code,
            "set_name": card.set_name,
            "collector_number": card.collector_number,
            "rarity": str(card.rarity),
            "color_identity": tutor.models.Color.to_string(card.color_identity),
            "cmc": float(card.cmc),
            "mana_cost": card.mana_cost,
            "type_line": card.type_line,
            "artist": card.artist,
            "oracle_text": card.oracle_text,
            "prices": {
                "usd": price(card.price_usd),
                "usd_foil": price(card.price_usd_foil),
                "eur": price(card.price_eur),
                "eur_foil": price(card.price_eur_foil),
                "tix": price(card.price_tix),
            },
        }

    def _card_copy(self, copy: tutor.models.CardCopy) -> dict:
        return {
            "card": self._card(copy.card),
            "foil": copy.foil,
            "collection": copy.collection,
        }

    def _deck_card(self, card: tutor.models.DeckCard) -> dict:
        return {
            "card": self._card(card.card),
            "quantity": card.quantity,
        }

    def _deck(self, deck: tutor.models.Deck) -> dict:
        return {
            "deck_id": deck.deck_id,
            "name": deck.name,
            "cards": [self._deck_card(card) for card in deck.cards],
        }


class OpenAPIRequestHandler(tornado_openapi3.handler.OpenAPIRequestHandler):
    @property
    def spec_dict(self):
        spec = getattr(self.application, "openapi_spec_dict", None)
        if not spec:
            version = importlib.metadata.version(__package__)
            spec = yaml.safe_load(self.render_string("openapi.yaml", version=version))
            del spec["servers"]
            setattr(self.application, "openapi_spec_dict", spec)
        return spec

    @property
    def spec(self):
        spec = getattr(self.application, "openapi_spec", None)
        if not spec:
            spec = super().spec
            setattr(self.application, "openapi_spec", spec)
        return spec

    def on_openapi_error(self, status_code, error):
        logger.info("OpenAPI Error: %s", error)
        super().on_openapi_error(status_code, error)


class RequestHandler(tornado.web.RequestHandler):
    def url(self, url: str) -> str:
        scheme_override = self.application.settings["scheme"]
        if not scheme_override:
            return url
        parts = urllib.parse.urlsplit(url)
        return urllib.parse.urlunsplit(
            (
                scheme_override,
                parts.netloc,
                parts.path,
                parts.query,
                parts.fragment,
            )
        )

    def set_links(self, **links) -> None:
        self.set_header(
            "Link",
            ", ".join(
                [f'<{self.url(url)}>; rel="{rel}"' for rel, url in links.items()]
            ),
        )

    @property
    def pool(self) -> psycopg_pool.AsyncConnectionPool:
        pool = getattr(self.application, "pool", None)
        if not pool:
            raise RuntimeError("Database pool not initialized")
        return pool


class StatusHandler(RequestHandler):
    async def get(self) -> None:
        self.set_header("Content-Type", "application/json")
        self.set_header("Access-Control-Allow-Origin", "*")
        try:
            pool = self.pool
        except RuntimeError:
            raise tornado.web.HTTPError(503)

        async with pool.connection() as conn:
            async with conn.cursor() as cursor:
                last_updated = await tutor.database.get_var(cursor, "last_update")

        self.write(
            json.dumps(
                {
                    "version": importlib.metadata.version(__package__),
                    "last_updated": str(last_updated),
                }
            )
        )


class SearchHandler(RequestHandler):
    async def get(self) -> None:
        async with self.pool.connection() as conn:
            async with conn.cursor() as cursor:
                query = self.get_argument("q", "")
                in_collection = self.get_argument("in_collection", None)
                page = max(1, int(self.get_argument("page", "1")))
                limit = int(self.get_argument("limit", "10"))
                sort_by = self.get_argument("sort_by", "rarity")
                search = tutor.search.search.parse(query)
                logger.debug(search)
                copies = await tutor.database.advanced_search(
                    cursor,
                    search,
                    limit=limit + 1,
                    offset=limit * (page - 1),
                    sort_by=sort_by,
                    in_collection=(
                        in_collection in ("yes", "true") if in_collection else None
                    ),
                )
        has_more = copies and len(copies) > limit
        self.set_header("Content-Type", "application/json")
        self.set_header("Access-Control-Allow-Origin", "*")
        links = {}
        if page > 1:
            links["prev"] = update_args(self.request.full_url(), page=page - 1)
        if has_more:
            links["next"] = update_args(self.request.full_url(), page=page + 1)
        self.set_links(**links)

        self.write(json.dumps(copies[:limit], cls=JSONEncoder))


class CollectionHandler(RequestHandler):
    async def get(self) -> None:
        async with self.pool.connection() as conn:
            async with conn.cursor() as cursor:
                self.set_header("Content-Type", "application/json")
                self.set_header("Access-Control-Allow-Origin", "*")
                self.write(
                    json.dumps(
                        await tutor.database.collection_stats(cursor), cls=JSONEncoder
                    )
                )


class DeckImporter:
    line_pattern = re.compile(r"^(?P<quantity>\d+)x? (?P<name>[^\]$]+)")

    def __init__(self) -> None:
        self.imported: int = 0
        self.last_chunk: bytes = b""

    def process(
        self, chunk: bytes = b""
    ) -> typing.Iterator[tutor.models.CardConstraint]:
        lines = (self.last_chunk + chunk).splitlines()
        self.last_chunk = lines.pop()
        for line in lines:
            if match := self.line_pattern.match(line.decode("utf8").strip()):
                for card in itertools.repeat(
                    tutor.models.CardConstraint(name=match.group("name")),
                    int(match.group("quantity")),
                ):
                    self.imported += 1
                    yield card


class DecksHandler(OpenAPIRequestHandler, RequestHandler):
    def initialize(self) -> None:
        self.importer = DeckImporter()
        self.deck_id = None

    async def get(self) -> None:
        page = max(1, int(self.get_argument("page", "1")))
        limit = int(self.get_argument("limit", "10"))
        async with self.pool.connection() as conn:
            async with conn.cursor() as cursor:
                decks = await tutor.database.get_decks(
                    cursor, limit=limit + 1, offset=limit * (page - 1)
                )
        has_more = decks and len(decks) > limit
        self.set_header("Content-Type", "application/json")
        self.set_header("Access-Control-Allow-Origin", "*")
        links = {}
        if page > 1:
            links["prev"] = update_args(self.request.full_url(), page=page - 1)
        if has_more:
            links["next"] = update_args(self.request.full_url(), page=page + 1)
        self.set_links(**links)
        self.write(json.dumps(decks, cls=JSONEncoder))

    async def post(self) -> None:
        self.set_header("Content-Type", "application/json")
        self.set_header("Access-Control-Allow-Origin", "*")
        logger.info("Validated: %s", self.validated)
        name = self.validated.body["name"]
        async with self.pool.connection() as conn:
            async with conn.cursor() as cursor:
                deck_id = await tutor.database.store_deck(cursor, name)
        if not deck_id:
            logger.error("Failed to store new deck")
            raise tornado.web.HTTPError(500)
        deck = tutor.models.Deck(deck_id, name, [])
        self.write(json.dumps(deck, cls=JSONEncoder))


@tornado.web.stream_request_body
class DeckHandler(RequestHandler):
    async def prepare(self) -> None:
        self.connection = await self.pool.getconn()
        self.deck_id = int(self.path_args[0])
        content_type = self.request.headers.get("Content-Type", "text/plain")
        if content_type == "text/plain":
            self.importer = DeckImporter()
        if self.request.method == "PUT":
            async with self.connection.cursor() as cursor:
                await tutor.database.clear_deck(cursor, self.deck_id)

    async def data_received(self, chunk) -> None:
        for card in self.importer.process(chunk):
            await self.add_card(card)

    async def add_card(self, card: tutor.models.CardConstraint) -> None:
        async with self.connection.cursor() as cursor:
            oracle_id = await tutor.database.oracle_id_by_name(cursor, card.name)
            if oracle_id:
                logger.debug("Adding card %s to deck %s", card.name, self.deck_id)
                await tutor.database.store_deck_card(cursor, self.deck_id, oracle_id, 1)

    def on_finish(self) -> None:
        IOLoop.current().add_callback(self.close_connection)

    async def close_connection(self) -> None:
        await self.pool.putconn(self.connection)

    async def get(self, deck_id) -> None:
        self.set_header("Content-Type", "application/json")
        self.set_header("Access-Control-Allow-Origin", "*")
        async with self.connection.cursor() as cursor:
            deck = await tutor.database.get_deck(cursor, deck_id)
        if not deck:
            raise tornado.web.HTTPError(404)
        self.write(json.dumps(deck, cls=JSONEncoder))

    async def post(self, deck_id) -> None:
        for card in self.importer.process():
            await self.add_card(card)
        async with self.connection.cursor() as cursor:
            deck = await tutor.database.get_deck(cursor, deck_id)
        await self.connection.commit()
        self.set_header("Content-Type", "application/json")
        self.set_header("Access-Control-Allow-Origin", "*")
        self.write(json.dumps(deck, cls=JSONEncoder))

    async def put(self, deck_id) -> None:
        for card in self.importer.process():
            await self.add_card(card)
        async with self.connection.cursor() as cursor:
            deck = await tutor.database.get_deck(cursor, deck_id)
        await self.connection.commit()
        self.set_header("Content-Type", "application/json")
        self.set_header("Access-Control-Allow-Origin", "*")
        self.write(json.dumps(deck, cls=JSONEncoder))

    async def delete(self, deck_id) -> None:
        async with self.pool.connection() as conn:
            async with conn.cursor() as cursor:
                await tutor.database.delete_deck(cursor, deck_id)
        self.set_status(204)


class DeckCandidatesHandler(OpenAPIRequestHandler, RequestHandler):
    async def get(self, deck_id: int) -> None:
        self.set_header("Content-Type", "application/json")
        self.set_header("Access-Control-Allow-Origin", "*")
        async with self.pool.connection() as conn:
            async with conn.cursor() as cursor:
                candidates = await tutor.database.deck_candidates(cursor, deck_id)
        self.write(json.dumps(candidates, cls=JSONEncoder))


class TemplateHandler(RequestHandler):
    def initialize(
        self,
        path: str,
        content_type: str = "application/octet-stream",
        vars: typing.Optional[dict] = None,
    ):
        self.path = path
        self.content_type = content_type
        self.vars = vars if vars else dict()

    async def get(self) -> None:
        self.set_header("Content-Type", self.content_type)
        self.render(self.path, **self.vars)


class StaticFileHandler(tornado.web.StaticFileHandler):
    @classmethod
    def get_absolute_path(cls, root: str, path: str) -> str:
        # Rewrite paths to load the index
        if any(
            re.match(pattern, path)
            for pattern in [
                r"^collection$",
                r"^decks$",
                r"^decks/.*",
            ]
        ):
            path = "index.html"
        return tornado.web.StaticFileHandler.get_absolute_path(root, path)


class Application(tornado.web.Application):
    def __init__(self, **settings):
        version = importlib.metadata.version(__package__)
        template_path = importlib.resources.files(__package__) / "templates"
        settings["template_path"] = str(template_path)

        if settings.get("autoreload"):
            import tornado.autoreload

            tornado.autoreload.watch(str(template_path / "openapi.yaml"))

        paths = [
            (
                r"/api",
                TemplateHandler,
                {"path": "index.html", "content_type": "text/html"},
            ),
            (
                r"/api/openapi.yaml",
                TemplateHandler,
                {
                    "path": "openapi.yaml",
                    "content_type": "application/x-yaml; charset=utf-8",
                    "vars": {"version": version},
                },
            ),
            (r"/api/search", SearchHandler),
            (r"/api/status", StatusHandler),
            (r"/api/collection", CollectionHandler),
            (r"/api/decks", DecksHandler),
            (r"/api/decks/([1-9][0-9]*)", DeckHandler),
            (r"/api/decks/([1-9][0-9]*)/candidates", DeckCandidatesHandler),
        ]
        if static_path := settings.get("static"):
            paths.extend(
                [
                    (
                        rf"/(.*)",
                        StaticFileHandler,
                        {"path": static_path, "default_filename": "index.html"},
                    ),
                ]
            )
        IOLoop.current().add_callback(self.async_init)
        super().__init__(paths, **settings)

    async def async_init(self):
        self.pool = psycopg_pool.AsyncConnectionPool(self.settings["database"])
