module App exposing (main)

import Browser
import Browser.Dom
import Browser.Events
import Browser.Navigation
import Card
import Color
import Dict exposing (Dict)
import Element as E
import Element.Background as Background
import Element.Border as Border
import Element.Events as Events
import Element.Font as Font
import Element.Input as Input
import Html.Events
import Http
import Json.Decode
import Maybe.Extra
import Pages.Collection
import Pages.DeckEditor
import Pages.DeckList
import Paginated
import Route
import Spinner
import Symbol
import Task
import UI
import Url
import Url.Builder
import Url.Parser exposing ((</>), (<?>))
import Url.Parser.Query


type alias Model =
    { navigationKey : Browser.Navigation.Key
    , url : Url.Url
    , viewport : UI.Dimensions
    , device : E.Device
    , route : Maybe Route.Route
    , symbols : Dict String Symbol.Symbol
    , page : Page
    }


type Msg
    = UrlChanged Url.Url
    | ViewportChanged UI.Dimensions
    | LinkClicked Browser.UrlRequest
    | CollectionMsg Pages.Collection.Msg
    | DeckListMsg Pages.DeckList.Msg
    | DeckEditorMsg Pages.DeckEditor.Msg
    | GotSymbology (Result Http.Error (List Symbol.Symbol))
    | SpinnerMsg Spinner.Msg


type Page
    = NotFound
    | Collection Pages.Collection.Model
    | DeckList Pages.DeckList.Model
    | DeckEditor Pages.DeckEditor.Model


init : Json.Decode.Value -> Url.Url -> Browser.Navigation.Key -> ( Model, Cmd Msg )
init _ url key =
    let
        viewport =
            { width = 1280, height = 720 }

        device =
            E.classifyDevice viewport

        route =
            Route.fromUrl url

        symbols =
            Dict.empty

        initWith : (pageModel -> Page) -> (pageMsg -> Msg) -> ( pageModel, Cmd pageMsg ) -> ( Page, Cmd Msg )
        initWith pageType pageMsg ( subModel, subCmd ) =
            ( pageType subModel, Cmd.map pageMsg subCmd )

        ( page, pageCmd ) =
            case route of
                Just Route.Collection ->
                    initWith Collection CollectionMsg <|
                        Pages.Collection.init key url device symbols

                Just Route.DeckList ->
                    initWith DeckList DeckListMsg <|
                        Pages.DeckList.init key url device

                Just (Route.Deck deckId) ->
                    initWith DeckEditor DeckEditorMsg <|
                        Pages.DeckEditor.init key url device symbols deckId

                _ ->
                    ( NotFound, Cmd.none )
    in
    ( { navigationKey = key
      , url = url
      , viewport = viewport
      , device = device
      , route = route
      , symbols = symbols
      , page = page
      }
    , Cmd.batch
        [ UI.getViewport ViewportChanged
        , getSymbology
        , pageCmd
        ]
    )


updateWith : (pageModel -> Page) -> (pageMsg -> Msg) -> Model -> ( pageModel, Cmd pageMsg ) -> ( Model, Cmd Msg )
updateWith pageType pageMsg model ( subModel, subCmd ) =
    ( { model | page = pageType subModel }, Cmd.map pageMsg subCmd )


update : Msg -> Model -> ( Model, Cmd Msg )
update msg model =
    case ( msg, model.page ) of
        ( UrlChanged url, _ ) ->
            let
                oldRoute =
                    model.route

                newRoute =
                    Route.fromUrl url

                ( newModel, cmds ) =
                    case ( model.page, newRoute ) of
                        ( Collection pageModel, Just Route.Collection ) ->
                            Pages.Collection.update (Pages.Collection.UrlChanged url) pageModel
                                |> updateWith Collection CollectionMsg model

                        ( _, Just Route.Collection ) ->
                            Pages.Collection.init model.navigationKey url model.device model.symbols
                                |> updateWith Collection CollectionMsg model

                        ( _, Just Route.DeckList ) ->
                            Pages.DeckList.init model.navigationKey url model.device
                                |> updateWith DeckList DeckListMsg model

                        ( _, Just (Route.Deck deckId) ) ->
                            Pages.DeckEditor.init model.navigationKey url model.device model.symbols deckId
                                |> updateWith DeckEditor DeckEditorMsg model

                        _ ->
                            ( { model | page = NotFound }, Cmd.none )
            in
            ( { newModel | route = Route.fromUrl url }, cmds )

        ( LinkClicked urlRequest, _ ) ->
            case urlRequest of
                Browser.Internal url ->
                    ( model, Browser.Navigation.pushUrl model.navigationKey (Url.toString url) )

                Browser.External url ->
                    ( model, Browser.Navigation.load url )

        ( ViewportChanged viewport, _ ) ->
            ( { model
                | viewport = viewport
                , device = E.classifyDevice viewport
              }
            , Cmd.none
            )

        ( GotSymbology (Ok symbols), _ ) ->
            let
                table =
                    symbols
                        |> List.map (\s -> ( s.symbol, s ))
                        |> Dict.fromList

                ( newModel, cmd ) =
                    case model.page of
                        Collection pageModel ->
                            Pages.Collection.update (Pages.Collection.GotSymbols table) pageModel
                                |> updateWith Collection CollectionMsg model

                        DeckEditor pageModel ->
                            Pages.DeckEditor.update (Pages.DeckEditor.GotSymbols table) pageModel
                                |> updateWith DeckEditor DeckEditorMsg model

                        _ ->
                            ( model, Cmd.none )
            in
            ( { newModel
                | symbols = table
              }
            , cmd
            )

        ( SpinnerMsg spinnerMsg, Collection pageModel ) ->
            Pages.Collection.update (Pages.Collection.SpinnerMsg spinnerMsg) pageModel
                |> updateWith Collection CollectionMsg model

        ( CollectionMsg pageMsg, Collection pageModel ) ->
            Pages.Collection.update pageMsg pageModel
                |> updateWith Collection CollectionMsg model

        ( DeckListMsg pageMsg, DeckList pageModel ) ->
            Pages.DeckList.update pageMsg pageModel
                |> updateWith DeckList DeckListMsg model

        ( DeckEditorMsg pageMsg, DeckEditor pageModel ) ->
            Pages.DeckEditor.update pageMsg pageModel
                |> updateWith DeckEditor DeckEditorMsg model

        ( _, _ ) ->
            ( model, Cmd.none )


navBar : Model -> E.Element Msg
navBar model =
    let
        navLink : Route.Route -> String -> E.Element Msg
        navLink route text =
            E.link
                [ E.pointer
                , E.padding 10
                , Font.center
                , Background.color <|
                    if Just route == model.route then
                        UI.colors.primary

                    else
                        UI.colors.background
                , E.mouseOver [ Background.color UI.colors.primary ]
                ]
                { url = Route.toUrl route, label = E.text text }
    in
    E.row
        [ E.padding 10
        , E.spacing 10
        , E.width E.fill
        , Background.color UI.colors.navBar
        , Font.color UI.colors.text
        ]
        [ E.el [ E.width <| E.fillPortion 1 ] <| E.text "Tutor"
        , E.row [ E.width <| E.fillPortion 4, E.spacing 3 ]
            [ navLink Route.Collection "Collection"
            , navLink Route.DeckList "Decks"
            ]
        ]


view : Model -> Browser.Document Msg
view model =
    let
        viewPage page =
            case page of
                Collection pageModel ->
                    E.map CollectionMsg <| Pages.Collection.view pageModel

                DeckList pageModel ->
                    E.map DeckListMsg <| Pages.DeckList.view pageModel

                DeckEditor pageModel ->
                    E.map DeckEditorMsg <| Pages.DeckEditor.view pageModel

                NotFound ->
                    E.column
                        [ E.width E.fill
                        , E.height E.fill
                        , E.padding 20
                        , E.spacing 20
                        ]
                        [ E.el
                            [ Font.color UI.colors.title
                            , Font.size 60
                            , E.centerX
                            ]
                          <|
                            E.text "404 Not Found"
                        , E.el
                            [ Font.color UI.colors.text
                            , E.centerX
                            ]
                          <|
                            E.text "The page you requested could not be found."
                        ]
    in
    { title = "Tutor"
    , body =
        [ E.layout
            [ Background.color UI.colors.background
            , E.height E.fill
            ]
          <|
            E.column
                [ E.width E.fill
                , E.height E.fill
                ]
                [ navBar model
                , viewPage model.page
                ]
        ]
    }


subscriptions : Model -> Sub Msg
subscriptions model =
    let
        global =
            [ Browser.Events.onResize
                (\w h -> ViewportChanged { width = w, height = h })
            ]
    in
    case model.page of
        Collection pageModel ->
            Sub.batch (Sub.map CollectionMsg (Pages.Collection.subscriptions pageModel) :: global)

        DeckList pageModel ->
            Sub.batch (Sub.map DeckListMsg (Pages.DeckList.subscriptions pageModel) :: global)

        DeckEditor pageModel ->
            Sub.batch (Sub.map DeckEditorMsg (Pages.DeckEditor.subscriptions pageModel) :: global)

        _ ->
            Sub.batch global


getSymbology : Cmd Msg
getSymbology =
    Http.get
        { url = Url.Builder.crossOrigin "https://api.scryfall.com" [ "symbology" ] []
        , expect =
            Http.expectJson GotSymbology <|
                Json.Decode.at [ "data" ] <|
                    Json.Decode.list Symbol.decode
        }


main : Program Json.Decode.Value Model Msg
main =
    Browser.application
        { init = init
        , onUrlChange = UrlChanged
        , onUrlRequest = LinkClicked
        , view = view
        , update = update
        , subscriptions = subscriptions
        }
