module Pages.Collection exposing (..)

import Browser
import Browser.Dom
import Browser.Events
import Browser.Navigation
import Card
import Color
import Dict
import Dropdown
import Element as E
import Element.Background as Background
import Element.Border as Border
import Element.Events as Events
import Element.Font as Font
import Element.Input as Input
import Html.Events
import Http
import Json.Decode
import Json.Decode.Pipeline as JDP
import Maybe.Extra
import Paginated
import Spinner
import Symbol
import Task
import UI
import Url
import Url.Builder
import Url.Parser exposing ((</>), (<?>))
import Url.Parser.Query


type alias Statistics =
    { cards : Int
    , sets : Int
    , value : Float
    }


decodeStatistics : Json.Decode.Decoder Statistics
decodeStatistics =
    Json.Decode.succeed Statistics
        |> JDP.required "cards" Json.Decode.int
        |> JDP.required "sets" Json.Decode.int
        |> JDP.required "value" Json.Decode.float


type SortBy
    = PriceDescending
    | RarityDescending
    | CreatedDescending


type alias Criteria =
    { query : String
    , sortBy : SortBy
    , ownedOnly : Bool
    }


type alias Model =
    { navigationKey : Browser.Navigation.Key
    , url : Url.Url
    , device : E.Device
    , spinner : Spinner.Model
    , sortDropdown : Dropdown.State SortBy
    , criteria : Criteria
    , cardPage : CardPage
    , activeCard : Maybe Card.Copy
    , collectionStatistics : Maybe Statistics
    , symbols : Symbol.Table
    }


type Msg
    = UrlChanged Url.Url
    | ViewportChanged UI.Dimensions
    | LinkClicked Browser.UrlRequest
    | SpinnerMsg Spinner.Msg
    | DropdownMsg (Dropdown.Msg SortBy)
    | SortSelected (Maybe SortBy)
    | UpdateCriteria CriteriaMsg
    | Search
    | GetPage Url.Url
    | GotStatistics (Result Http.Error Statistics)
    | FoundCards (Result Http.Error (Paginated.Page Card.Copy))
    | ShowCardDetails Card.Copy
    | ClearCardDetails
    | GotSymbols Symbol.Table


type CriteriaMsg
    = UpdateName String
    | UpdateSortBy SortBy
    | UpdateOwnedOnly Bool


type CardPage
    = Ready (Paginated.Page Card.Copy)
    | Loading (Paginated.Page Card.Copy)
    | Failed


toLoading : CardPage -> CardPage
toLoading cardPage =
    case cardPage of
        Ready page ->
            Loading page

        Loading page ->
            Loading page

        Failed ->
            Loading Paginated.empty


searchQuery : Criteria -> List Url.Builder.QueryParameter
searchQuery criteria =
    let
        toString : SortBy -> String
        toString sortBy =
            case sortBy of
                PriceDescending ->
                    "price"

                RarityDescending ->
                    "rarity"

                CreatedDescending ->
                    "created"
    in
    [ Url.Builder.string "q" criteria.query
    , Url.Builder.string "sort_by" (toString criteria.sortBy)
    , Url.Builder.string "in_collection"
        (if criteria.ownedOnly then
            "yes"

         else
            "no"
        )
    ]


search : Criteria -> Cmd Msg
search criteria =
    loadPage <|
        Url.Builder.absolute
            [ "api", "search" ]
            (Url.Builder.int "limit" 18 :: searchQuery criteria)


loadPage : String -> Cmd Msg
loadPage url =
    Http.get
        { url = url
        , expect = Paginated.expectJson FoundCards Card.decodeCopy
        }


getCollectionStatistics : Cmd Msg
getCollectionStatistics =
    Http.get
        { url = Url.Builder.absolute [ "api", "collection" ] []
        , expect = Http.expectJson GotStatistics decodeStatistics
        }


parseUrl : Url.Parser.Parser (Criteria -> a) a
parseUrl =
    let
        query =
            Url.Parser.Query.string "q"
                |> Url.Parser.Query.map (Maybe.withDefault "")

        sortBy =
            Url.Parser.Query.enum "sort_by"
                (Dict.fromList
                    [ ( "rarity", RarityDescending )
                    , ( "price", PriceDescending )
                    , ( "created", CreatedDescending )
                    ]
                )
                |> Url.Parser.Query.map (Maybe.withDefault PriceDescending)

        inCollection =
            Url.Parser.Query.enum "in_collection"
                (Dict.fromList
                    [ ( "true", True )
                    , ( "false", False )
                    , ( "yes", True )
                    , ( "no", False )
                    ]
                )
                |> Url.Parser.Query.map (Maybe.withDefault True)
    in
    Url.Parser.oneOf
        [ Url.Parser.top <?> Url.Parser.Query.map3 Criteria query sortBy inCollection
        , Url.Parser.s "collection" <?> Url.Parser.Query.map3 Criteria query sortBy inCollection
        ]


criteriaFromUrl : Url.Url -> Criteria
criteriaFromUrl url =
    let
        emptyCriteria =
            { query = "", sortBy = PriceDescending, ownedOnly = True }
    in
    Url.Parser.parse parseUrl url
        |> Maybe.withDefault emptyCriteria


init : Browser.Navigation.Key -> Url.Url -> E.Device -> Symbol.Table -> ( Model, Cmd Msg )
init key url device symbols =
    let
        criteria =
            criteriaFromUrl url
    in
    ( { navigationKey = key
      , url = url
      , device = device
      , spinner = Spinner.init
      , sortDropdown = Dropdown.init "sort-dropdown"
      , criteria = criteria
      , cardPage = Loading Paginated.empty
      , activeCard = Nothing
      , collectionStatistics = Nothing
      , symbols = symbols
      }
    , Cmd.batch
        [ UI.getViewport ViewportChanged
        , search criteria
        , getCollectionStatistics
        ]
    )


updateCriteria : CriteriaMsg -> Criteria -> Criteria
updateCriteria msg model =
    case msg of
        UpdateName text ->
            { model | query = text }

        UpdateSortBy column ->
            { model | sortBy = column }

        UpdateOwnedOnly value ->
            { model | ownedOnly = value }


update : Msg -> Model -> ( Model, Cmd Msg )
update msg model =
    case msg of
        UrlChanged url ->
            let
                criteria =
                    criteriaFromUrl url
            in
            ( { model | criteria = criteria }, search criteria )

        LinkClicked _ ->
            ( model, Cmd.none )

        ViewportChanged viewport ->
            ( { model
                | device = E.classifyDevice viewport
              }
            , Cmd.none
            )

        SpinnerMsg msg_ ->
            ( { model | spinner = Spinner.update msg_ model.spinner }, Cmd.none )

        DropdownMsg msg_ ->
            let
                ( updated, cmd ) =
                    Dropdown.update sortDropdown msg_ model model.sortDropdown
            in
            ( { model | sortDropdown = updated }, cmd )

        SortSelected sortBy ->
            let
                newValue : SortBy
                newValue =
                    Maybe.withDefault PriceDescending sortBy
            in
            update (UpdateCriteria (UpdateSortBy newValue)) model

        UpdateCriteria msg_ ->
            let
                newCriteria =
                    updateCriteria msg_ model.criteria
            in
            case msg_ of
                UpdateName _ ->
                    ( { model | criteria = newCriteria }
                    , Cmd.none
                    )

                UpdateSortBy _ ->
                    update Search { model | criteria = newCriteria }

                UpdateOwnedOnly _ ->
                    update Search { model | criteria = newCriteria }

        Search ->
            ( { model
                | cardPage = toLoading model.cardPage
                , activeCard = Nothing
              }
            , Cmd.batch
                [ Browser.Navigation.pushUrl model.navigationKey <|
                    Url.Builder.relative [] (searchQuery model.criteria)
                ]
            )

        GetPage url ->
            ( { model | cardPage = toLoading model.cardPage }, loadPage (Url.toString url) )

        GotStatistics (Ok statistics) ->
            ( { model | collectionStatistics = Just statistics }, Cmd.none )

        GotStatistics (Err _) ->
            ( model, Cmd.none )

        FoundCards (Ok cardPage) ->
            ( { model | cardPage = Ready cardPage }, Cmd.none )

        FoundCards (Err _) ->
            ( model, Cmd.none )

        ShowCardDetails card ->
            ( { model | activeCard = Just card }, Cmd.none )

        ClearCardDetails ->
            ( { model | activeCard = Nothing }, Cmd.none )

        GotSymbols symbols ->
            ( { model | symbols = symbols }, Cmd.none )


searchBar : Model -> E.Element Msg
searchBar model =
    let
        alignment =
            if UI.isMobile model.device then
                E.column

            else
                E.row
    in
    alignment
        [ E.padding 10
        , E.spacing 10
        , E.width E.fill
        , Background.color UI.colors.navBar
        ]
        [ E.row [ E.spacing 10, E.width E.fill ]
            [ Input.text
                [ onEnter Search
                , Background.color UI.colors.background
                , Font.color UI.colors.text
                , E.width (E.fill |> E.minimum 150)
                ]
                { onChange = UpdateCriteria << UpdateName
                , text = model.criteria.query
                , placeholder = Nothing
                , label = Input.labelHidden "Search Input"
                }
            , Input.button
                [ Background.color UI.colors.primary
                , Font.color UI.colors.text
                , Border.rounded 10
                , E.padding 10
                ]
                { onPress = Just Search
                , label = E.text "Search"
                }
            ]
        , E.row [ E.spacing 10 ]
            [ Dropdown.view sortDropdown model model.sortDropdown ]
        ]


sortDropdown =
    let
        toString s =
            case s of
                PriceDescending ->
                    "Price DESC"

                RarityDescending ->
                    "Rarity DESC"

                CreatedDescending ->
                    "Created DESC"

        selectAttrs =
            [ Background.color UI.colors.secondary
            , Border.width 1
            , Border.rounded 5
            , Border.color UI.colors.background
            , E.paddingXY 16 8
            , E.spacing 10
            , E.width E.fill
            , E.pointer
            ]

        listAttrs =
            [ Background.color UI.colors.secondary
            , Border.width 1
            , Border.roundEach { topLeft = 0, topRight = 0, bottomLeft = 5, bottomRight = 5 }
            , E.width E.fill
            , E.spacing 5
            , E.pointer
            ]

        itemToPrompt sortBy =
            E.text ("Sort By " ++ toString sortBy)

        itemToElement selected highlighted sortBy =
            E.el
                [ E.padding 8
                , E.spacing 10
                , E.width E.fill
                , E.mouseOver [ Background.color UI.colors.primary ]
                ]
            <|
                E.text <|
                    toString sortBy
    in
    Dropdown.basic
        { itemsFromModel =
            always
                [ PriceDescending
                , RarityDescending
                , CreatedDescending
                ]
        , selectionFromModel = .criteria >> .sortBy >> Just
        , dropdownMsg = DropdownMsg
        , onSelectMsg = SortSelected
        , itemToPrompt = itemToPrompt
        , itemToElement = itemToElement
        }
        |> Dropdown.withSelectAttributes selectAttrs
        |> Dropdown.withListAttributes listAttrs


viewCardBrowser : Model -> E.Element Msg
viewCardBrowser model =
    let
        navButton text maybeUrl =
            case maybeUrl of
                Just url ->
                    Input.button
                        [ E.height (E.px 30)
                        , E.width E.fill
                        , Background.color UI.colors.primary
                        , Border.rounded 5
                        , Font.color UI.colors.text
                        , Font.center
                        ]
                        { label = E.text text, onPress = Just (GetPage url) }

                Nothing ->
                    E.el [ E.width E.fill ] E.none

        cards cardPage =
            let
                attrs =
                    if UI.isMobile model.device then
                        [ E.width E.fill
                        , E.height <| E.fillPortion 3
                        ]

                    else
                        [ E.width <| E.fillPortion 3
                        , E.height E.fill
                        ]

                copyRow copy =
                    let
                        cardAttrs =
                            [ E.pointer, Events.onClick <| ShowCardDetails copy ]
                                ++ (if model.activeCard == Just copy then
                                        [ Background.color UI.colors.selected ]

                                    else
                                        []
                                   )
                    in
                    UI.cardRow
                        { foil = copy.foil
                        , subtitle = String.join " " [ copy.collection ]
                        }
                        cardAttrs
                        model.symbols
                        copy.card
            in
            E.column attrs
                [ E.row
                    [ E.spacing 5
                    , E.padding 5
                    , E.width E.fill
                    ]
                    [ navButton "<-" cardPage.prev
                    , navButton "->" cardPage.next
                    ]
                , E.column
                    [ E.width E.fill
                    , E.height E.fill
                    , E.scrollbarY
                    ]
                  <|
                    List.map copyRow cardPage.values
                ]
    in
    case model.cardPage of
        Failed ->
            E.none

        Loading cardPage ->
            E.el
                [ E.height E.fill
                , E.centerX
                ]
            <|
                E.html <|
                    Spinner.view UI.manaSpinner model.spinner

        Ready cardPage ->
            cards cardPage


onEnter : msg -> E.Attribute msg
onEnter msg =
    E.htmlAttribute
        (Html.Events.on "keyup"
            (Json.Decode.field "key" Json.Decode.string
                |> Json.Decode.andThen
                    (\key ->
                        if key == "Enter" then
                            Json.Decode.succeed msg

                        else
                            Json.Decode.fail "Not the enter key"
                    )
            )
        )


viewCardDetails model =
    let
        viewCard : Card.Copy -> E.Element Msg
        viewCard copy =
            let
                maxDimensions =
                    { width = 480, height = 680 }
            in
            E.el
                [ Border.rounded 10
                , E.clip
                , E.width (E.fill |> E.maximum maxDimensions.width)
                ]
            <|
                E.image
                    [ E.width E.fill
                    , E.behindContent <|
                        E.html <|
                            Spinner.view UI.manaSpinner model.spinner
                    ]
                    { src =
                        Url.Builder.crossOrigin "https://api.scryfall.com"
                            [ "cards", copy.card.scryfallId ]
                            [ Url.Builder.string "format" "image"
                            , Url.Builder.string "version" "border_crop"
                            ]
                    , description = copy.card.name
                    }

        prices card isFoil =
            Maybe.Extra.values
                [ Maybe.map (\usd -> { currency = "usd", amount = usd }) <|
                    if isFoil then
                        Maybe.Extra.or card.prices.usd_foil card.prices.usd

                    else
                        Maybe.Extra.or card.prices.usd card.prices.usd_foil
                , Maybe.map (\eur -> { currency = "eur", amount = eur }) <|
                    if isFoil then
                        Maybe.Extra.or card.prices.eur_foil card.prices.eur

                    else
                        Maybe.Extra.or card.prices.eur card.prices.eur_foil
                , Maybe.map (\tix -> { currency = "tix", amount = tix }) card.prices.tix
                ]

        cardDetails copy =
            let
                cardImage =
                    if UI.isMobile model.device then
                        E.none

                    else
                        E.row [ E.width E.fill ]
                            [ E.el [ E.width <| E.fillPortion 1 ] E.none
                            , E.el [ E.width <| E.fillPortion 5 ] <|
                                viewCard copy
                            , E.el [ E.width <| E.fillPortion 1 ] E.none
                            ]
            in
            E.column
                [ E.spacing 20
                , E.padding 10
                , E.width E.fill
                , E.height E.fill
                , Font.color UI.colors.text
                ]
                [ cardImage
                , E.paragraph [ Font.heavy, Font.size 24, Font.center, Font.color UI.colors.title ]
                    [ E.text copy.card.name ]
                , E.row [ E.spacing 5, E.centerX ] <|
                    List.map UI.priceBadge (prices copy.card copy.foil)
                , E.paragraph [Font.size 16, Font.italic, Font.center] [ UI.text "Artist: ", UI.text copy.card.artist ]
                , E.paragraph [] <| [ Symbol.text model.symbols 20 copy.card.manaCost, UI.text <| " " ++ copy.card.typeLine ]
                , E.column [ E.spacing 10 ] <|
                    List.map (Symbol.text model.symbols 16) (String.lines copy.card.oracleText)
                ]

        closedetails =
            Input.button
                [ E.height (E.px 30)
                , E.width E.fill
                , Background.color UI.colors.secondary
                , Border.rounded 5
                , Font.color UI.colors.text
                , Font.center
                ]
                { label = E.text "Close", onPress = Just ClearCardDetails }
    in
    if UI.isMobile model.device then
        E.column
            [ E.width E.fill
            , E.height E.fill
            ]
            [ model.activeCard
                |> Maybe.map cardDetails
                |> Maybe.withDefault E.none
            , closedetails
            ]

    else
        E.el
            [ E.alignTop
            , E.width <| E.fillPortion 1
            , E.height E.fill
            , E.scrollbarY
            , Background.color UI.colors.sidebar
            ]
            (model.activeCard
                |> Maybe.map cardDetails
                |> Maybe.withDefault
                    E.none
            )


view : Model -> E.Element Msg
view model =
    let
        footer =
            case model.collectionStatistics of
                Just statistics ->
                    Just <|
                        E.el [ E.centerY, Font.size 16, Font.italic ] <|
                            E.text <|
                                String.concat
                                    [ String.fromInt statistics.cards
                                    , " cards in collection spanning "
                                    , String.fromInt statistics.sets
                                    , " sets (Estimated value: $"
                                    , String.fromFloat statistics.value
                                    , ")"
                                    ]

                Nothing ->
                    Nothing
    in
    UI.layout model.device
        { toolbar = Just <| searchBar model
        , sidebar =
            if UI.isMobile model.device then
                Maybe.map (\_ -> viewCardDetails model) model.activeCard

            else
                Just <| viewCardDetails model
        , content = viewCardBrowser model
        , footer = footer
        }


subscriptions : Model -> Sub Msg
subscriptions model =
    Sub.batch
        [ Browser.Events.onResize
            (\w h -> ViewportChanged { width = w, height = h })
        , Sub.map SpinnerMsg Spinner.subscription
        ]
