module Pages.DeckList exposing (..)

import Browser
import Browser.Events
import Browser.Navigation
import Deck
import Element as E
import Element.Background as Background
import Element.Events as Events
import Element.Font as Font
import Http
import Paginated
import Route
import Spinner
import UI
import Url
import Url.Builder


type alias Model =
    { navigationKey : Browser.Navigation.Key
    , url : Url.Url
    , device : E.Device
    , spinner : Spinner.Model
    , deckPage : Deckpage
    }


type Msg
    = ViewportChanged UI.Dimensions
    | SpinnerMsg Spinner.Msg
    | GotDecks (Result Http.Error (Paginated.Page Deck.Deck))
    | DeckClicked Int


type Deckpage
    = Ready (Paginated.Page Deck.Deck)
    | Loading (Paginated.Page Deck.Deck)
    | Failed


init : Browser.Navigation.Key -> Url.Url -> E.Device -> ( Model, Cmd Msg )
init key url device =
    ( { navigationKey = key
      , url = url
      , device = device
      , spinner = Spinner.init
      , deckPage = Loading Paginated.empty
      }
    , getDecks
    )


update : Msg -> Model -> ( Model, Cmd Msg )
update msg model =
    case msg of
        ViewportChanged viewport ->
            ( { model
                | device = E.classifyDevice viewport
              }
            , Cmd.none
            )

        SpinnerMsg msg_ ->
            ( { model | spinner = Spinner.update msg_ model.spinner }, Cmd.none )

        GotDecks (Ok deckPage) ->
            ( { model | deckPage = Ready deckPage }, Cmd.none )

        GotDecks (Err _) ->
            ( { model | deckPage = Failed }, Cmd.none )

        DeckClicked deckId ->
            ( model, Browser.Navigation.pushUrl model.navigationKey (Route.toUrl (Route.Deck deckId)) )


view : Model -> E.Element Msg
view model =
    case model.deckPage of
        Failed ->
            E.none

        Loading _ ->
            E.el [ E.height E.fill, E.centerX ] <|
                E.html <|
                    Spinner.view UI.manaSpinner
                        model.spinner

        Ready deckPage ->
            if deckPage == Paginated.empty then
                E.column [ E.padding 40, E.centerX, Font.italic ] [ UI.text "No decks yet" ]

            else
                let
                    deckRow deck =
                        E.row
                            [ E.width E.fill
                            , E.padding 10
                            , E.spacing 3
                            , E.pointer
                            , E.mouseOver [ Background.color UI.colors.hover ]
                            , Events.onClick <| DeckClicked deck.id
                            ]
                            [ UI.title deck.name ]
                in
                UI.layout model.device
                    { toolbar = Nothing
                    , sidebar = Nothing
                    , content =
                        E.column [ E.width E.fill, E.centerX ] <| List.map deckRow deckPage.values
                    , footer = Nothing
                    }


subscriptions : Model -> Sub Msg
subscriptions model =
    Sub.batch
        [ Browser.Events.onResize
            (\w h -> ViewportChanged { width = w, height = h })
        , Sub.map SpinnerMsg Spinner.subscription
        ]


getDecks : Cmd Msg
getDecks =
    loadPage <|
        Url.Builder.absolute
            [ "api", "decks" ]
            [ Url.Builder.int "limit" 18 ]


loadPage : String -> Cmd Msg
loadPage url =
    Http.get
        { url = url
        , expect = Paginated.expectJson GotDecks Deck.decode
        }
