module Symbol exposing (Symbol, Table, Text, decode, text)

import Dict exposing (Dict)
import Element as E
import Element.Background as Background
import Element.Border as Border
import Element.Font as Font
import Json.Decode
import Json.Decode.Pipeline as JDP
import Parser as P exposing ((|.), (|=))


type alias Symbol =
    { symbol : String
    , svgURI : String
    , english : String
    }


type alias Table =
    Dict String Symbol


type Text
    = Text String
    | Inline Symbol


decode : Json.Decode.Decoder Symbol
decode =
    Json.Decode.succeed Symbol
        |> JDP.required "symbol" Json.Decode.string
        |> JDP.required "svg_uri" Json.Decode.string
        |> JDP.required "english" Json.Decode.string


symbolTextParser : Dict String Symbol -> P.Parser (List Text)
symbolTextParser available =
    let
        lookupSymbol : String -> Text
        lookupSymbol string =
            Dict.get string available
                |> Maybe.map Inline
                |> Maybe.withDefault (Text "")

        textParser : P.Parser Text
        textParser =
            P.chompUntilEndOr "{"
                |> P.getChompedString
                |> P.map Text

        symbolStringParser : P.Parser String
        symbolStringParser =
            P.getChompedString <|
                P.succeed ()
                    |. P.symbol "{"
                    |. P.chompWhile (\c -> c /= '}')
                    |. P.symbol "}"

        symbolParser : P.Parser Text
        symbolParser =
            symbolStringParser
                |> P.map lookupSymbol
    in
    P.loop []
        (\reversedParts ->
            P.oneOf
                [ P.end
                    |> P.map (\_ -> P.Done (List.reverse reversedParts))
                , symbolParser
                    |> P.map (\part -> P.Loop (part :: reversedParts))
                , textParser
                    |> P.map (\part -> P.Loop (part :: reversedParts))
                ]
        )


toElement : Int -> Text -> E.Element msg
toElement size symbolText =
    let
        symbol : Symbol -> E.Element msg
        symbol s =
            E.image
                [ E.width <| E.px size
                , E.height <| E.px size
                ]
                { src = s.svgURI
                , description = s.english
                }
    in
    case symbolText of
        Text s ->
            E.text s

        Inline s ->
            symbol s


text : Dict String Symbol -> Int -> String -> E.Element msg
text available size string =
    E.paragraph [ E.spacing (size // 10), Font.size size ] <|
        List.map (toElement size) <|
            (P.run (symbolTextParser available) string |> Result.withDefault [])
